/*
 * File:   main.c
 * Author: Daniela Mangione
 *
 * 2014 - Controllo di una sospensione cardanica
 */

#define FCY 39613750
#define BRGVAL 85 			//Value UxBRG for Baud Rate = 15.156 bps
#define SLAVE_ADDRESS 0xA6 
#define PG 3.14159265

#include <stdio.h>
#include <string.h>
#include <xc.h>
#include <p33FJ128MC802.h>
#include <p33Fxxxx.h>
#include <libpic30.h>
#include <i2c.h>
#include <math.h>
#include "servo_out.h"
/*
 *
 */
#define INPUT 1
#define OUTPUT 0

/*Accelerometro*/
#define BW_RATE 0x2C //Data Rate and power control mode
#define PWR_CTL 0x2D //Power control
//ASSI
#define DATAX0 0x32	 //Address Register
#define DATAY0 0x34 
#define DATAZ0 0x36 
//
/*
 *
 */
//dsPIC33FJ128MC802 Configuration Bit Settings
 #pragma config FNOSC = FRCPLL           // Oscillator Mode (Fast RC Oscillator with PLL (FRCPLL))
 #pragma config IESO = ON                // Internal External Switch Over Mode (Start-up device with FRC, then automatically switch to user-selected oscillator source when ready)
 #pragma config POSCMD = NONE            // Primary Oscillator Source (Primary Oscillator Disabled)
 #pragma config OSCIOFNC = ON            // OSC2 Pin Function (OSC2 pin has digital I/O function)
 #pragma config IOL1WAY = ON             // Peripheral Pin Select Configuration (Allow Only One Re-configuration)
 #pragma config FCKSM = CSECME           // Clock Switching and Monitor (Both Clock Switching and Fail-Safe Clock Monitor are enabled)
 #pragma config FWDTEN = OFF             // Watchdog Timer Enable (Watchdog timer enabled/disabled by user software)
 #pragma config FPWRT = PWR128           // POR Timer Value (128ms)
 #pragma config ALTI2C = ON              // Alternate I2C  pins (I2C mapped to SDA1/SCL1 pins)
 #pragma config ICS = PGD1               // Comm Channel Select (Communicate on PGC1/EMUC1 and PGD1/EMUD1)
 #pragma config JTAGEN = OFF             // JTAG Port Enable (JTAG is Disabled)
/*
 *
 */
int x, y, z;
float ax, ay, az; //in g
int data_register_0, data_register_1;
//MASTER
/*Assert Start condition on SDAx e SCLx*/
void start()
{
    I2C1CONbits.SEN = 1; 			//begin start condition. Baud Rate Generator start
     //Hardware: S = 1, P = 0;
    while(I2C1CONbits.SEN !=0); 	//wait for end start condition
    IFS1bits.MI2C1IF = 0;		   //reset interrupt flag
}
void restart()
{
    I2C1CONbits.RSEN = 1; //begin start condition. Baud Rate Generator start
    //Hardware: S = 1, P = 0;
    while(I2C1CONbits.RSEN !=0); //wait for end start condition
    IFS1bits.MI2C1IF = 0; //reset interrupt flag
}
/*Generate a Stop condition on SDAx and SCLx.*/
void stop()
{
    I2C1CONbits.PEN = 1; //stop condition
    while(I2C1CONbits.PEN !=0);
    IFS1bits.MI2C1IF =0; //reset interrupt flag bit
    __delay_ms(1);
}
int data_send(unsigned char data) //master to slave.
{
    //Sending Data to slave device
    I2C1TRN =  data; //0 -write
    while(I2C1STATbits.TRSTAT!=0);
    //while(IFS1bits.MI2C1IF != 1);
    IFS1bits.MI2C1IF =0;
    return(I2C1STATbits.ACKSTAT ==0);
}
int write_register(unsigned char dev_addr, unsigned char reg_num, unsigned char value)
{
   start();
   if (data_send(dev_addr) == 0) {
      // errore!
      stop();
      return 0;
   }
   if (data_send(reg_num) == 0) {
      // errore!
      stop();
      return 0;
   }
   if (data_send(value) == 0) {
      // errore!
      stop();
      return 0;
   }
   stop();
   return 1; // tutto e' andato bene!
}
/*Enable master reception to receive serial memory data*/
int data_receive(int a_n)
{
    I2C1CONbits.RCEN = 1; //Enables Receive mode for I2C. Hardware cleared
    while(I2C1STATbits.RBF !=1); //1 = Receive complete; I2CxRCV register is full
    LATAbits.LATA0= 1;
    int data= I2C1RCV;
    //Genera ACK/NACK
    if(a_n == 0){
        //send ACK
        I2C1CONbits.ACKDT = 0; //send ACK;
    }
    else if(a_n ==1){
        //send NACK
        I2C1CONbits.ACKDT = 1; //send NACK;
    }
    I2C1CONbits.ACKEN = 1; //Initiates a master Acknowledge event
    while(I2C1CONbits.ACKEN !=0);
    IFS1bits.MI2C1IF = 0;
    return data;
}
int read_register(unsigned char dev_addr, unsigned char reg_num) {
    //STEP 1 - START
    start();
    //STEP 2-3 - send address + ACK
    if (data_send(dev_addr) == 0) { //write address
      // errore!
      stop();
      return 0;
   }
   //STEP 4-5 - send register number + ACK
   if (data_send(reg_num) == 0) { //write register
      // errore!
      stop();
      return 0;
   }
   //STEP 6 - RESTART
    restart();
   //STEP 7 - Writing in Read-Mode
    data_send(dev_addr|1);
    //STEP 8-9 - send ACK
    data_register_0 = data_receive(0);
    //STEP 10-11 - send NACK
    data_register_1 = data_receive(1);
    //STEP 12 - STOP
    stop();
    return 1;
}
/*
 *
 */
int main() {
    float pitch, roll;
    int val;
    //i2c: pin SCL e SDA config
    TRISBbits.TRISB14 = INPUT; //SCL as input
    TRISBbits.TRISB15 = INPUT; //SDA as input
    LATBbits.LATB14= 0;
    LATBbits.LATB15= 0;

    /*
    TRISAbits.TRISA0 = OUTPUT;
    LATAbits.LATA0 = 0;
	*/
	
    //oscillator config
    //FOSC = Fin * M/(N1*N2) = 7.37 *43/(2*2) =  79.22 Mhz
    CLKDIVbits.FRCDIV  = 0b000;         //FRC divided by 1 (default)
    CLKDIVbits.PLLPRE  = 0b00000;       //N1 =2  Input divided by 2 (default)
    CLKDIVbits.PLLPOST = 0b00;          //N2 =2    Output divided by 2
    PLLFBDbits.PLLDIV  = 0b000101001;   //PLLDIV = 41;  M = PLLDIV + 2 = 43

    /*configurazione UART*/
    TRISBbits.TRISB3 = INPUT;    // RX PIN set to input
    TRISBbits.TRISB2 = OUTPUT;   //TX PIN set to output
    RPINR18bits.U1RXR = 0b00011; //RX pin. RP3
    RPOR1bits.RP2R = 0b00011;    //TX pin. RP2

    //setup UART trasmitter
    U1MODEbits.BRGH = 1; //Low Speed mode
    U1STAbits.UTXEN = 1; //UARTx transmitter enabled; UxTX pin is controlled by UARTx
    //Setup UART receiver
    U1STAbits.URXISEL = 0; // Interrupt after one RX character is received;
    //Impostazionni Baud Rate
    U1MODEbits.ABAUD = 0; // Auto-Baud Disabled
    U1BRG = BRGVAL; //UxBRG = [FCY/(4*BaudRate)] -1 -> [39613750/(4*115156)] -1  = 85
    U1MODEbits.UEN = 0b10; //UxTX, UxRX, UxCTSand UxRTS pins are enabled and used
    //Altre configurazioni
    U1MODEbits.IREN = 0; //IRDA disabled
    U1MODEbits.WAKE = 1; //wake-up disabled. (=0 enabled on start bit detected during sleep)
    U1MODEbits.LPBACK = 0; //loopback mode disabled
    U1MODEbits.USIDL = 0; //continue operation in idle mode
    U1MODEbits.URXINV = 1; //polarit- idle state = 0
    U1MODEbits.PDSEL = 0b00; //No Parity, 8 data bits
    U1MODEbits.STSEL = 0; // 1 Stop bit
    //Attivazione modulo
    U1MODEbits.UARTEN = 1;  //Enable UART

    /*Setup I2C*/
    //master
    I2C1STATbits.I2COV = 0; //Receive Overflow Flag bit
    IFS1bits.MI2C1IF = 0; //General master interrupt events
    I2C1BRG = 196; //Baud Rate di circa 200 kHz
    I2C1CONbits.I2CEN = 1; //enables the i2c1 module and configure SDA1 e SCL1 pin as serial port pins
    //inizializzazione slave
    write_register(SLAVE_ADDRESS, BW_RATE, 0xc); //BW_RATE = 0xc
    write_register(SLAVE_ADDRESS, PWR_CTL, 8); //PWR_CTL = 8  - POWER ON!
    
	//init
    init_servo();

    for(;;) {
        read_register(SLAVE_ADDRESS, DATAX0);
        x = data_register_0 + data_register_1*256;
        ax= (float)x/256;
        read_register(SLAVE_ADDRESS, DATAY0);
        y = data_register_0 + data_register_1*256;
        ay= (float)y/256;
        read_register(SLAVE_ADDRESS, DATAZ0);
        z = data_register_0 + data_register_1*256;
        az= (float)z/256;
	
        pitch = atan2(-ay, az);
        roll = atan2(ax, (sqrt(pow(ay, 2)+ pow(az, 2))));
        //conversion grad to rad 
        pitch= (pitch*180)/PG;
        roll= (roll*180)/PG;

        //roll
        if(roll >= 90 ) {
             set_servo(3, SERVO_HIGH); //roll a +90
        }
        if(roll == 0){
            set_servo(3, SERVO_HIGH/2); //roll a +0
        }
        if(roll <=-90) {
            set_servo(3, 1); //roll a -90
        }
        else {
            //90: HIGH = roll: X
            val = (SERVO_HIGH* (roll+90))/180;
            set_servo(3, val); //roll
        }

        //pitch
        if(pitch >= 90 ) {
             set_servo(4, SERVO_HIGH); //pitch a +90
        }
        if(pitch == 0){
            set_servo(4, SERVO_HIGH/2); //pitch a +0
        }
        if(pitch <=-90) {
            set_servo(4, 1); //pitch a -90
        }
        else {
            //90: HIGH = pitch: X
            val = (SERVO_HIGH* (pitch+90))/180;
            set_servo(4, val); //pitch
        }
        //__delay_ms(500);
        //print data on screen
        printf("x= %f  y= %f  z= %f pitch= %f roll= %f\n", ax, ay, az, pitch, roll); //STAMPA IN G
    }

    return (0);
}
