/*
 * servo_out.c
 */

#include <p33FJ128MC802.h>
#include <libpic30.h>
#include <stdio.h>

#include "servo_out.h"

int servo_2_val = 0;

#define SERVO_2_PORT       PORTBbits.RB14
#define SERVO_2_PORT_DIR   TRISBbits.TRISB14

#define PWM_PERIOD_2_5      1510
// ~2.5 millisecond period (400 Hz) at 77.385 MHz clock, prescaler = 64

#define PWM_PERIOD_5        3020
// ~5 millisecond period (200 Hz) at 77.385 MHz clock, prescaler = 64

#define PWM_PERIOD_10       6040
// ~10 millisecond period (100 Hz) at 77.385 MHz clock, prescaler = 64

#define PWM_PERIOD_20       12080
// ~20 millisecond period (50 Hz) at 77.385 MHz clock, prescaler = 64

int servo_out[4];

void __attribute__((__interrupt__,__no_auto_psv__)) _T2Interrupt(void)
{
    SERVO_2_PORT = 0;
    T2CONbits.TON = 0;
    IFS0bits.T2IF = 0;
}

void __attribute__((__interrupt__,__no_auto_psv__)) _MPWM1Interrupt(void)
{
    if (servo_2_val == 0) {
        SERVO_2_PORT = 0;
    }
    else {
        SERVO_2_PORT = 1;
        PR2 = servo_2_val;
        TMR2 = 0;
        T2CONbits.TON = 1;
        IFS0bits.T2IF = 0;
    }
    IFS3bits.PWM1IF = 0;
}

void init_servo(void)
{
    PTPER = PWM_PERIOD_20;

    PTCONbits.PTCKPS = 0b11;	// prescaler = 64
    PTCONbits.PTOPS = 0;	// no postscaler
    PTCONbits.PTMOD = 0b00;	// free run

    PWMCON1bits.PMOD1 = 1 ; // independent PWM mode
    PWMCON1bits.PMOD2 = 1 ;
    PWMCON1bits.PMOD3 = 1 ;

    PWMCON1bits.PEN1L = 1 ; // low pins used as PWM
    PWMCON1bits.PEN2L = 1 ;
    PWMCON1bits.PEN3L = 1 ;

    PWMCON1bits.PEN1H = 0 ; // high pins used as digital I/O
    PWMCON1bits.PEN2H = 0 ;
    PWMCON1bits.PEN3H = 0 ;

    PDC1 = PDC2 = PDC3 = 0;

    ODCBbits.ODCB15 = 0; // no open-drain
    ODCBbits.ODCB14 = 0; // no open-drain
    ODCBbits.ODCB13 = 0; // no open-drain
    ODCBbits.ODCB11 = 0; // no open-drain

    TRISBbits.TRISB15 = 0;
    TRISBbits.TRISB14 = 0;
    TRISBbits.TRISB13 = 0;
    TRISBbits.TRISB11 = 0;

    IFS3bits.PWM1IF = 0;
    IEC3bits.PWM1IE = 1;
    IPC14bits.PWM1IP = 3; // priority 3

    // -------------- timer 2 for 4th software-based pwm channel

    T2CONbits.TON = 0;
    T2CONbits.TSIDL = 1;
    T2CONbits.TGATE = 0;
    T2CONbits.TCKPS = 0b10; // x 64 prescaler
    T2CONbits.T32 = 0; // 16 bit timer

    TMR2 = 0;
    servo_2_val = 0;

    PR2 = PWM_PERIOD_20;
    IFS0bits.T2IF = 0;
    IPC1bits.T2IP = 3; // priority 3
    IEC0bits.T2IE = 1;

    SERVO_2_PORT = 0;
    SERVO_2_PORT_DIR = 0;

    set_servo(1, 0);
    set_servo(2, 0);
    set_servo(3, 0);
    set_servo(4, 0);

    PTCONbits.PTEN = 1; 	// turn on the PWM
}

void stop_servo(void)
{
    set_servo(1, 0);
    set_servo(2, 0);
    set_servo(3, 0);
    set_servo(4, 0);
}

void set_servo(int servo_num, int value)
{
    servo_out[servo_num - 1] = value;

    if (value == 0) {
        switch (servo_num) {
        case 1: PDC1 = 0; break;
        case 2: servo_2_val = 0; break;
        case 3: PDC2 = 0; break;
        case 4: PDC3 = 0; break;
        }
    }
    else {
        value += SERVO_LIMIT_LOW;

        if (value < SERVO_LIMIT_LOW) value = SERVO_LIMIT_LOW;
        if (value > SERVO_LIMIT_HIGH) value = SERVO_LIMIT_HIGH;

        switch (servo_num) {
        case 1: PDC1 = value; break;
        case 2: servo_2_val = value / 2; break;
        case 3: PDC2 = value; break;
        case 4: PDC3 = value; break;
        }
    }
}

